-- Noise Module Tree Specification
-- A class to allow a user to specify a chain of noise modules as a table of function declarations in a concise fashion.

require 'Class'
--dofile("gradient_fractal_scaleoffsets.lua")




NoiseModuleTree=class(function(a,declaration)
	a.moduletable={}
	a.seedablemodules={}
	if declaration~=nil then
		a:parseDeclaration(declaration)
	end
end)

function NoiseModuleTree:getFunction(name)
	return self.moduletable[name]
end

-- Functions to parse module parameters

-- Function to verify the existence and type a numeric parameter, with a warning and a default
function NoiseModuleTree:verifyNumberWarn(module, name, default)
	local v=module[name]
	if v==nil or type(v)~="number" then
		print("Warning: Parameter "..name.." of module "..module.name.." not correctly specified. Defaulting to "..default)
		return default
	end
	
	print("Parameter "..name.." of module "..module.name..": "..v)
	return v
end

-- Function to verify the existence and type of a numeric parameter, with an error
function NoiseModuleTree:verifyNumberError(module,name)
	local v=module[name]
	if v==nil or type(v)~="number" then
		print("Error: Parameter "..name.." of module "..module.name.." not specified.")
		return false
	end
	
	print("Parameter "..name.." of module "..module.name..": "..v)
	return v
end

function NoiseModuleTree:verifyNumber(module,name,default)
	if default==nil then return self:verifyNumberError(module,name)
	else return self:verifyNumberWarn(module,name,default)
	end
end

-- Function to verify a boolean parameter, with warn and default
function NoiseModuleTree:verifyBoolean(module,name,default)
	local v=module[name]
	if v~=true and v~=false then
		print("Warning: Boolean parameter "..name.." of module "..module.name.." not specified. Defaulting to "..tostring(default))
		return(default)
	end
	
	print("Parameter "..name.." of module "..module.name..": "..tostring(v))
	return v
end

-- Function to verify a string parameter
function NoiseModuleTree:verifyStringError(module, name)
	local v=module[name]
	if v==nil or type(v)~="string" then
		print("Error: String parameter "..name.." of module "..module.name.." not specified.")
		return false
	end
	
	print("Parameter "..name.." of module "..module.name..": "..v)
	return v
end

function NoiseModuleTree:verifyStringWarn(module,name,default)
	local v=module[name]
	if v==nil or type(v)~="string" then
		print("Warning: String parameter "..name.." of module "..module.name.." not specified. Defaulting to "..default)
		return default
	end
	
	print("Parameter "..name.." of module "..module.name..": "..v)
	return v
end

function NoiseModuleTree:verifyString(module,name,default)
	if default==nil then return self:verifyStringError(module,name)
	else return self:verifyStringWarn(module,name,default)
	end
end

-- Function to verify a source module
function NoiseModuleTree:verifyModuleError(module, name)
	local v=module[name]
	if v==nil or type(v)~="string" then
		print("Error: Module parameter "..name.." of module "..module.name.." not specified.")
		return false
	end
	
	local m=self.moduletable[v]
	if m==nil then
		print("Error: Module parameter "..m.." specified for module "..module.name.." does not exist.")
		return false
	end
	
	print("Parameter "..name.." of module "..module.name..": "..v)
	return m
end

function NoiseModuleTree:parseDeclaration(declaration)
	local i
	for i=1,#declaration,1 do
		if self:parseModule(declaration[i])==false then
			-- Error condition
			print("Aborting")
			return
		end
	end
	print("Declaration parsed. "..(#declaration).." modules.")
end

function NoiseModuleTree:setSeedFunction(name,seed)
	local f=self:getFunction(name)
	if f then f:setSeed(seed) end
end

function NoiseModuleTree:setSeedAll(startseed)
	local i,j
	local k=0
	for i,j in pairs(self.moduletable) do
		print("Seeding "..i)
		j:setSeed(startseed+k*10)
		k=k+1
	end
end

function NoiseModuleTree:parseModule(module)
	if module==nil then
		print("Error: nil module table passed to parseModule")
		return false
	end
	
	-- Check name
	if module.name==nil then
		print("Error: module name not specified")
		return false
	end
	
	if self.moduletable[module.name] ~= nil then
		print("Error: Module "..module.name.." already exists.")
		return false
	end
	
	-- Check type
	
	if module.type==nil then
		print("Error: module type not specified.")
		return false
	end
	
	-- The name of the module type directly references the name of a function to parse the module
	if self[module.type]==nil then
		print("Error: module type "..module.type.." unsupported.")
		return false
	end
	
	return self[module.type](self,module)
end

function NoiseModuleTree:gatherMultipleSources(module,func)
	-- Parse the source_n entries of module, check and obtain the modules and set the sources
	for i=0,MaxSources-1,1 do
		local s="source_"..i
		if module[s]~=nil then
			if self.moduletable[module[s]]==nil then
				print("Error: Module "..module[s].." specified as source to module "..module.name.." does not exist.")
				return false
			end
			if module[s]==module.name then
				print("Error: Module "..module.name.." specified as source for itself; will result in infinite loop condition.")
				return false
			end
			func:setImplicitSource(i,self.moduletable[module[s]])
			print("Module "..module[s].."set as source "..i.." of module "..module.name)
		end
	end
end

function NoiseModuleTree:gatherMultipleSourcesGreyscale(module, func)
	-- Parse the source_n entries of module, check and obtain the modules and set the sources
	for i=0,MaxSources-1,1 do
		local s="source_"..i
		if module[s]~=nil then
			if self.moduletable[module[s]]==nil then
				print("Error: Module "..module[s].." specified as source to module "..module.name.." does not exist.")
				return false
			end
			if module[s]==module.name then
				print("Error: Module "..module.name.." specified as source for itself; will result in infinite loop condition.")
				return false
			end
			func:setGreyscaleSource(i,self.moduletable[module[s]])
			print("Module "..module[s].."set as source "..i.." of module "..module.name)
		end
	end
end

function NoiseModuleTree:constant(module)
	local constant=1
	if module.constant == nil then
		print("Warning: Constant value not specified in module "..module.name..". Defaulting to 1.")
	else
		constant=module.constant
	end
		
	self.moduletable[module.name]=CImplicitConstant(constant)
	print("Module "..module.name.."(constant,"..constant..") created.")
end

function NoiseModuleTree:combiner(module)
	local type
	if module.combiner_type=="ADD" then
		type=ECT_ADD
	elseif module.combiner_type=="MULTIPLY" then
		type=ECT_MULT
	elseif module.combiner_type=="MAX" then
		type=ECT_MAX
	elseif module.combiner_type=="MIN" then
		type=ECT_MIN
	elseif module.combiner_type=="AVERAGE" then
		type=ECT_AVG
	elseif module.combiner_type==nil then
		print("Warning: combiner_type specified for module "..module.name.." is nil; defaulting to ADD")
		type=ECT_ADD
	else
		print("Warning: combiner_type specified for module "..module.name.." is of unknown type; defaulting to ADD")
		type=ECT_ADD
	end
	
	self.moduletable[module.name]=CImplicitCombiner(type)
	self:gatherMultipleSources(module,self.moduletable[module.name])
	print("Module "..module.name.."(combiner,"..type..") created.")
	return true
end

function NoiseModuleTree:scaleoffset(module)
	local scale,offset=module.scale,module.offset
	if scale==nil then
		print("Warning: scale not specified for module "..module.name..". Defaulting to 1")
		scale=1
	end
	
	if offset==nil then
		print("Warning: offset not specified for module "..module.name..". Defaulting to 0")
		offset=0
	end
	
	self.moduletable[module.name]=CImplicitScaleOffset(scale,offset)
	
	if module.source==nil then
		print("Error: Source module not specified for module "..module.name)
		return false
	end
	
	if self.moduletable[module.source]==nil then
		print("Error: Source module "..module.source.." specified as source for "..module.name.." does not exist.")
		return false
	end
	
	self.moduletable[module.name]:setImplicitSource(self.moduletable[module.source])
	print("Module "..module.name.."(scaleoffset,"..scale..","..offset..") create.")
	return true
end

function NoiseModuleTree:turbulence(module)
	local f=CImplicitTurbulence()
	
	local mainsourcename=module.main_source
	if mainsourcename==nil or self.moduletable[mainsourcename]==nil then
		print("Error: Main source not specified for module "..module.name)
		return false
	end
	f:setImplicitSource(self.moduletable[mainsourcename])
	print("Module "..module.name.."(turbulence,"..mainsourcename.." created.")
	
	if module.x_axis_source ~= nil and self.moduletable[module.x_axis_source] ~= nil then
		f:setAxisSource(0,self.moduletable[module.x_axis_source])
		print("X Axis source set to "..module.x_axis_source)
	end
	if module.y_axis_source ~= nil and self.moduletable[module.y_axis_source] ~= nil then
		f:setAxisSource(1,self.moduletable[module.y_axis_source])
		print("Y Axis source set to "..module.y_axis_source)
	end
	if module.z_axis_source ~= nil and self.moduletable[module.z_axis_source] ~= nil then
		f:setAxisSource(2,self.moduletable[module.z_axis_source])
		print("Z Axis source set to "..module.z_axis_source)
	end
	if module.w_axis_source ~= nil and self.moduletable[module.w_axis_source] ~= nil then
		f:setAxisSource(3,self.moduletable[module.w_axis_source])
		print("W Axis source set to "..module.w_axis_source)
	end
	if module.u_axis_source ~= nil and self.moduletable[module.u_axis_source] ~= nil then
		f:setAxisSource(4,self.moduletable[module.u_axis_source])
		print("U Axis source set to "..module.u_axis_source)
	end
	if module.v_axis_source ~= nil and self.moduletable[module.v_axis_source] ~= nil then
		f:setAxisSource(5,self.moduletable[module.v_axis_source])
		print("V Axis source set to "..module.v_axis_source)
	end
	
	if module.x_power~=nil then
		f:setXPower(module.x_power)
		print("X Power set to "..module.x_power)
	end
	if module.y_power~=nil then
		f:setYPower(module.y_power)
		print("Y Power set to "..module.y_power)
	end
	if module.z_power~=nil then
		f:setZPower(module.z_power)
		print("Z Power set to "..module.z_power)
	end
	if module.w_power~=nil then
		f:setWPower(module.w_power)
		print("W Power set to "..module.w_power)
	end
	if module.u_power~=nil then
		f:setUPower(module.u_power)
		print("U Power set to "..module.u_power)
	end
	if module.v_power~=nil then
		f:setVPower(module.v_power)
		print("V Power set to "..module.v_power)
	end
	
	if module.mask_source~=nil and self.moduletable[module.mask_source]~=nil then
		f:setMaskSource(self.moduletable[module.mask_source])
		print("Mask source set to "..module.mask_source)
	end
	self.moduletable[module.name]=f
	
	return true
end

function NoiseModuleTree:clamp(module)
	local source=self:verifyModuleError(module, "source")
	if source==false then
		print("Error: Could invalid source for clamp module.")
		return false
	end
	
	local low=self:verifyNumber(module,"low",-1)
	local high=self:verifyNumber(module,"high",1)
	
	local f=CImplicitClamp(low,high)
	f:setImplicitSource(source)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (clamp) created.")
	return true
end

function NoiseModuleTree:select(module)
	local mainsource,lowsource,highsource=module.main_source, module.low_source, module.high_source
	if mainsource==nil or self.moduletable[mainsource]==nil then
		print("Error: Main source module for "..module.name.." not specified.")
		return false
	elseif lowsource==nil or self.moduletable[lowsource]==nil then
		print("Error: Low source module for "..module.name.." not specified.")
		return false
	elseif highsource==nil or self.moduletable[highsource]==nil  then
		print("Error: High source module for "..module.name.." not specified.")
	end
	
	local threshold,falloff=module.threshold,module.falloff
	if threshold==nil then
		print("Warning: Threshold for module "..module.name.." not specified; defaulting to 0")
		threshold=0
	end
	if falloff==nil then
		print("Warning: Falloff for module "..module.name.." not specified; defaulting to 0")
		falloff=0
	end
	
	local f=CImplicitSelect()
	f:setImplicitSource(0,self.moduletable[lowsource])
	f:setImplicitSource(1,self.moduletable[highsource])
	f:setImplicitSource(2,self.moduletable[mainsource])
	f:setThreshold(threshold)
	f:setFalloff(falloff)
	self.moduletable[module.name]=f
	print("Module "..module.name.."(select,"..threshold..","..falloff..") created")
	return true
end

function NoiseModuleTree:blend(module)
	local mainsource,lowsource,highsource=module.main_source, module.low_source, module.high_source
	if mainsource==nil or self.moduletable[mainsource]==nil then
		print("Error: Main source module for "..module.name.." not specified.")
		return false
	elseif lowsource==nil or self.moduletable[lowsource]==nil then
		print("Error: Low source module for "..module.name.." not specified.")
		return false
	elseif highsource==nil or self.moduletable[highsource]==nil  then
		print("Error: High source module for "..module.name.." not specified.")
	end
	
	local f=CImplicitBlend()
	f:setImplicitSource(0,self.moduletable[lowsource])
	f:setImplicitSource(1,self.moduletable[highsource])
	f:setImplicitSource(2,self.moduletable[mainsource])
	self.moduletable[module.name]=f
	print("Module "..module.name.."(blend) created")
	return true
end

function NoiseModuleTree:gradient(module)
	local x1,y1,z1,w1,u1,v1, x2,y2,z2,w2,u2,v2=0,0,0,0,0,0,0,0,0,0,0,0
	
	if module.x1~=nil then x1=module.x1 end
	if module.x2~=nil then x2=module.x2 end
	if module.y1~=nil then y1=module.y1 end
	if module.y2~=nil then y2=module.y2 end
	if module.z1~=nil then z1=module.z1 end
	if module.z2~=nil then z2=module.z2 end
	if module.w1~=nil then w1=module.w1 end
	if module.w2~=nil then w2=module.w2 end
	if module.u1~=nil then u1=module.u1 end
	if module.u2~=nil then u2=module.u2 end
	if module.v1~=nil then v1=module.v1 end
	if module.v2~=nil then v2=module.v2 end
	
	self.moduletable[module.name]=CImplicitGradient()
	self.moduletable[module.name]:setGradient(x1,x2,y1,y2,z1,z2,w1,w2,u1,u2,v1,v2)
	print("Module "..module.name.."(gradient) created.")
	return true
end

function NoiseModuleTree:fractal(module)
	local type,basis,interp,octaves,freq=EFT_FBM,EBT_GRADIENT,EIT_QUINTIC,8,1
	
	if module.type then
		if module.fractal_type=="FBM" then type=EFT_FBM print("Module "..module.name.." fractal type set to FBM")
		elseif module.fractal_type=="RIDGEDMULTI" then type=EFT_RIDGEDMULTI print("Module "..module.name.." fractal type set to RIDGEDMULTI")
		elseif module.fractal_type=="BILLOW" then type=EFT_BILLOW print("Module "..module.name.." fractal type set to BILLOW")
		elseif module.fractal_type=="MULTI" then type=EFT_MULTI print("Module "..module.name.." fractal type set to MULTI")
		elseif module.fractal_type=="HYBRIDMULTI" then type=EFT_HYBRIDMULTI print("Module "..module.name.." fractal type set to HYBRIDMULTI")
		else
			print("Error: Fractal type "..module.type.." not valid.")
			return false
		end
	else
		print("Warning: Fractal type for "..module.name.." not specified; defaulting to FBM.")
	end
	
	if module.basis_type then
	if module.basis_type=="VALUE" then basis=EBT_VALUE print("Module "..module.name.." basis type set to VALUE.")
	elseif module.basis_type=="GRADVAL" then basis=EBT_GRADVAL print("Module "..module.name.." basis type set to GRADVAL")
	elseif module.basis_type=="GRADIENT" then basis=EBT_GRADIENT print("Module "..module.name.." basis type set to GRADIENT")
	elseif module.basis_type=="WHITE" then basis=EBT_WHITE print("Module "..module.name.." basis type set to WHITE")
	elseif module.basis_type=="SIMPLEX" then basis=EBT_GRADIENT print("Warning: SIMPLEX basis not currently implemented; Module "..module.name.." basis type set to GRADIENT instead")
	else
		print("Error: Basis type "..module.basis_type.." not supported.")
		return false
	end
	end
	
	if module.interp_type then
	if module.interp_type=="NONE" then interp=EIT_NONE print("Module "..module.name.." interp set to NONE")
	elseif module.interp_type=="LINEAR" then interp=EIT_LINEAR print("Module "..module.name.." interp set to LINEAR")
	elseif module.interp_type=="CUBIC" then interp=EIT_CUBIC print("Module "..module.name.." interp set to CUBIC")
	elseif module.interp_type=="QUINTIC" then interp=EIT_QUINTIC print("Module "..module.name.." interp set to QUINTIC")
	else
		print("Error: Interp type "..module.interp_type.." not supported.")
		return false
	end
	end
	
	if module.num_octaves then
		octaves=module.num_octaves
		print("Module "..module.name.." octaves set to "..octaves)
	end
	
	if module.frequency then
		freq=module.frequency
		print("Module "..module.name.." frequency set to "..freq)
	end
	
	
	-- TODO: Implement setters for Lacunarity, Offset, Gain, H
	local f=CImplicitFractal(type,basis,interp)
	f:setNumOctaves(octaves)
	f:setFrequency(freq)
	
	-- Specify any overrides
	print("Module "..module.name..": Check for octave overrides...")
	self:gatherMultipleSources(module,f)
	
	if module.seed then f:setSeed(module.seed) end
	
	-- TODO: Implement a Usage hint to determine what dimensionality of normalizing factors to calculate
	f:calcNormalizingFactors_3D(10000)
	self.moduletable[module.name]=f
	print("Module "..module.name.."(fractal) created.")
	return true
end

function NoiseModuleTree:basisfunction(module)
	local basis_type=self:verifyNumber(module, "basis_type", 1)
	local interp_type=self:verifyNumber(module, "interp_type", 3)

	if basis_type<0 or basis_type>3 then
		print("Warning: Basis type of module "..module.name.." out of range. Defaulting to 1")
		basis_type=1
	end
	if interp_type<0 or interp_type>3 then
		print("Warning: Interp type of module "..module.name.." out of range. Defaulting to 3.")
		interp_type=3
	end
	
	local f=CImplicitBasisFunction(basis_type, interp_type)
	self.moduletable[module.name]=f
	
	print("Module "..module.name.." (basisfunction) created.")
	return true
end

function NoiseModuleTree:sphere(module)
	local cx=self:verifyNumber(module,"cx",0)
	local cy=self:verifyNumber(module,"cy",0)
	local cz=self:verifyNumber(module,"cz",0)
	local cw=self:verifyNumber(module,"cw",0)
	local cu=self:verifyNumber(module,"cu",0)
	local cv=self:verifyNumber(module,"cv",0)
	local radius=self:verifyNumber(module,"radius",0)
	
	local f=CImplicitSphere()
	f:setCenter(cx,cy,cz,cw,cu,cv)
	f:setRadius(radius)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (sphere) created.")
	return true
end


function NoiseModuleTree:bias(module)
	local bias=0.5
	
	if module.bias then
		bias=module.bias
		print("Module "..module.name.." bias set to "..bias)
	end
	
	self.moduletable[module.name]=CImplicitBias(bias)
	if module.source==nil or self.moduletable[module.source]==nil then
		print("Error: Source for module "..module.name.." not specified.")
		return false
	end
	self.moduletable[module.name]:setImplicitSource(self.moduletable[module.source])
	print("Module "..module.name.."(bias) created")
	return true
end

function NoiseModuleTree:gain(module)
	local gain=0.5
	
	if module.gain then
		gain=module.gain
		print("Module "..module.name.." gain set to "..gain)
	end
	
	self.moduletable[module.name]=CImplicitGain(gain)
	if module.source==nil or self.moduletable[module.source]==nil then
		print("Error: Source for module "..module.name.." not specified.")
		return false
	end
	self.moduletable[module.name]:setImplicitSource(self.moduletable[module.source])
	print("Module "..module.name.."(gain) created")
	return true
end

function NoiseModuleTree:clamprange(module)
	local low,high=-1,1
	
	if module.low then
		low=module.low
		print("Module "..module.name.." low clamp set to "..low)
	end
	if module.high then
		high=module.high
		print("Module "..module.name.." high clamp set to "..high)
	end
	
	self.moduletable[module.name]=CImplicitClampRange(low,high)
	if module.source==nil or self.moduletable[module.source]==nil then
		print("Error: Source for module "..module.name.." not specified.")
		return false
	end
	self.moduletable[module.name]:setImplicitSource(self.moduletable[module.source])
	print("Module "..module.name.."(clamprange) created.")
	return true
end

function NoiseModuleTree:curve(module)
	if module.points==nil or #module.points<2 then
		print("Error: Points array specified for module "..module.name.." nonexistent, or less than 2 entries in length.")
		return false
	end
	
	local f=CImplicitCurve()
	

	local i
	-- Set first point to -0.1 as boundary safeguard
	f:setControlPoint(-0.1,module.points[1])
	for i=1,#module.points,1 do
		local v=module.points[i]
		f:setControlPoint((i-1)/(#module.points - 1), v)
	end
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." created(curve)")
	return true
end

function NoiseModuleTree:scaledomain(module)
	local sx,sy,sz,sw,su,sv=1,1,1,1,1,1
	
	if module.scalex then sx=module.scalex end
	if module.scaley then sy=module.scaley end
	if module.scalez then sz=module.scalez end
	if module.scalew then sw=module.scalew end
	if module.scaleu then su=module.scaleu end
	if module.scalev then sv=module.scalev end
	
	self.moduletable[module.name]=CImplicitScaleDomain()
	self.moduletable[module.name]:setTransform(sx,sy,sz,sw,su,sv)
	if module.source==nil or self.moduletable[module.source]==nil then
		print("Error: Source for module "..module.name.." not specified.")
		return false
	end
	self.moduletable[module.name]:setImplicitSource(self.moduletable[module.source])
	print("Module "..module.name.."(scaledomain,"..sx..","..sy..","..sz..","..sw..","..su..","..sv.." created.")
	return true
end



-----------------------------------------------------------------------------
-- Greyscale modules
function NoiseModuleTree:verifyGetSource(sourcename)
	if sourcename==nil or self.moduletable[sourcename]==nil then return false
	else return self.moduletable[sourcename]
	end
end

function NoiseModuleTree:greyscale_scaleoffset(module)
	local source=self:verifyModuleError(module, "source")
	if source==false then return false end
	
	local scale=self:verifyNumber(module,"scale",1)
	local offset=self:verifyNumber(module,"offset",0)
	
	local f=CGreyscaleScaleOffset(scale,offset)
	f:setGreyscaleSource(source)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (greyscale_scaleoffset) created.")
	return true
end

function NoiseModuleTree:greyscale_implicitadapter(module)
	local source=self:verifyModuleError(module, "source")
	if source==false then return false end
	
	local mapx1=self:verifyNumber(module, "mapx1", 0)
	local mapx2=self:verifyNumber(module, "mapx2", 1)
	local mapy1=self:verifyNumber(module, "mapy1", 0)
	local mapy2=self:verifyNumber(module, "mapy2", 1)
	local mapz1=self:verifyNumber(module, "mapz1", 0)
	local mapz2=self:verifyNumber(module, "mapz2", 1)
	
	local seamlessmode=self:verifyNumber(module, "seamlessmode", ESM_NONE)
	
	local loopx1,loopy1,loopz1,loopx2,loopy2,loopz2=0,0,0,1,1,1
	if seamlessmode~=0 then
		loopx1=self:verifyNumber(module, "loopx1", 0)
		loopy1=self:verifyNumber(module, "loopy1", 0)
		loopz1=self:verifyNumber(module, "loopz1", 0)
		loopx2=self:verifyNumber(module, "loopx2", 1)
		loopy2=self:verifyNumber(module, "loopy2", 1)
		loopz2=self:verifyNumber(module, "loopz2", 1)
	end
	
	local zdepth=self:verifyNumber(module, "zdepth", 0)
	
	local f=CGreyscaleImplicitAdapter()
	f:setImplicitSource(source)
	f:setSeamlessMode(seamlessmode)
	
	f:setXMapRange(mapx1,mapx2)
	f:setYMapRange(mapy1,mapy2)
	f:setZMapRange(mapz1,mapz2)
	f:setXLoopRange(loopx1,loopx2)
	f:setYLoopRange(loopy1,loopy2)
	f:setZLoopRange(loopz1,loopz2)
	f:setZDepth(zdepth)
	
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (Implicit Adapter) created.")
	return true
end

function NoiseModuleTree:greyscale_bias(module)
	local bias=self:verifyNumber(module,"bias",0.5)
	local source=self:verifyModuleError(module,"source")
	if source==false then return false end
	
	local f=CGreyscaleBias()
	f:setBias(bias)
	f:setSource(source)
	self.moduletable[module.name]=f
	
	print("Module "..module.name.." (Greyscale Bias) created.")
	return true
end

function NoiseModuleTree:greyscale_blend(module)
	local mainsource=self:verifyModuleError(module, "main_source")
	local lowsource=self:verifyModuleError(module, "low_source")
	local highsource=self:verifyModuleError(module, "high_source")
	
	if mainsource==false or lowsource==false or highsource==false then return false end
	
	local f=CGreyscaleBlend()
	f:setGreyscaleSource(0,lowsource)
	f:setGreyscaleSource(1,highsource)
	f:setGreyscaleSource(2,mainsource)
	self.moduletable[module.name]=f
	print("Module "..module.name.."(greyscale_blend) created")
	return true
end

function NoiseModuleTree:greyscale_select(module)
	local mainsource=self:verifyModuleError(module,"main_source")
	local lowsource=self:verifyModuleError(module, "low_source")
	local highsource=self:verifyModuleError(module, "high_source")
	if mainsource==false or lowsource==false or highsource==false then return false end
	
	local threshold=self:verifyNumber(module, "threshold", 0)
	local falloff=self:verifyNumber(module, "falloff", 0)
	
	
	local f=CGreyscaleSelect()
	f:setGreyscaleSource(0,lowsource)
	f:setGreyscaleSource(1,highsource)
	f:setGreyscaleSource(2,mainsource)
	f:setThreshold(threshold)
	f:setFalloff(falloff)
	self.moduletable[module.name]=f
	print("Module "..module.name.."(greyscale_select,"..threshold..","..falloff..") created")
	return true
end

function NoiseModuleTree:greyscale_blur(module)
	local mainsource=self:verifyModuleError(module, "source")
	if mainsource==false then return false end
	
	local size=self:verifyNumber(module, "size")
	if size==false then return false end
	if size<=0 then print("Error: Blur size <=0") return false end
	
	local seamless=self:verifyBoolean(module, "seamless", false)
	
	local f=CGreyscaleBlur()
	f:setGreyscaleSource(mainsource)
	f:setSize(size)
	f:setSeamless(seamless)
	self.moduletable[module.name]=f
	print("Module "..module.name.." (greyscale_blur, "..size..","..tostring(seamless)..") created.")
	return true
end

function NoiseModuleTree:greyscale_buffer(module)
	local mainsource=self:verifyString(module, "source")
	if mainsource==false then return false end
	
	local f=CGreyscaleBuffer()
	f:loadImage(mainsource)
	self.moduletable[module.name]=f
	print("Module "..module.name.." (greyscale_buffer, "..mainsource..") created.")
	return true
end

function NoiseModuleTree:greyscale_bumpmap(module)
	local mainsource=self:verifyModuleError(module, "source")
	if mainsource==false then return false end
	
	local light, sharpness, seamless=module.light
	if light==nil then
		print("Error: Light vector for module "..module.name.." not specified.")
		return false
	end
	
	local sharpness=self:verifyNumber(module, "sharpness", 1)
	local seamless=self:verifyBoolean(module, "seamless", false)
	
	local f=CGreyscaleBumpMap()
	f:setGreyscaleSource(0,mainsource)
	f:setSharpness(sharpness)
	f:setSeamless(seamless)
	f:setLight(light[1], light[2], light[3])
	self.moduletable[module.name]=f
	print("Module "..module.name.." (greyscale_bumpmap, "..sharpness..","..tostring(seamless)..") created.")
	return true
end


function NoiseModuleTree:greyscale_cellular(module)
	local mapx1=self:verifyNumber(module,"mapx1",0)
	local mapy1=self:verifyNumber(module,"mapy1",0)
	local mapx2=self:verifyNumber(module,"mapx2",1)
	local mapy2=self:verifyNumber(module,"mapy2",1)
	
	local loopx1,loopy1,loopx2,loopy2=0,0,1,1
	
	if mapx1==false or mapy1==false or mapx2==false or mapy2==false then
		return false
	end
	
	local seamless=self:verifyBoolean(module, "seamless", false)
	if seamless==true then
		loopx1=self:verifyNumber(module,"loopx1", 0)
		loopy1=self:verifyNumber(module,"loopy1", 0)
		loopx2=self:verifyNumber(module,"loopx2", 1)
		loopy2=self:verifyNumber(module,"loopy2", 1)
	end
	
	local idoutput=self:verifyBoolean(module,"idoutput",false)
	local f1=self:verifyNumber(module, "f1", 0)
	local f2=self:verifyNumber(module, "f2", 0)
	local f3=self:verifyNumber(module, "f3", 0)
	local f4=self:verifyNumber(module, "f4", 0)
	
	local numpoints=self:verifyNumber(module, "numpoints", 10)
	
	local f=CGreyscaleCellular()
	f:setMapRanges(mapx1,mapx2,mapy1,mapy2)
	if seamless==true then
		f:setSeamless(seamless)
		f:setLoopRanges(loopx1,loopx2,loopy1,loopy2)
	end
	
	f:setIDOutput(idoutput)
	f:setCoefficients(f1,f2,f3,f4)
	self.moduletable[module.name]=f
	print("Module "..module.name.." (greyscale_cellular) created.")
	return true
end

function NoiseModuleTree:greyscale_combiner(module)
	local type=self:verifyNumber(module, "combiner_type", 0)
	if type<0 or type>4 then
		print("Warning: type parameter for module "..module.name.." out of range. Defaulting to 0.")
		type=0
	end
	
	local f=CGreyscaleCombiner(type)
	self:gatherMultipleSourcesGreyscale(module,f)
	self.moduletable[module.name]=f
	
	print("Module "..module.name.."(combiner,"..type..") created.")
	return true
end

function NoiseModuleTree:greyscale_constant(module)
	local constant=self:verifyNumber(module,"constant",0)
	
	local f=CGreyscaleConstant(constant)
	self.moduletable[module.name]=f
	print("Module "..module.name.." (constant, "..constant..") created.")
	return true
end

function NoiseModuleTree:greyscale_erosion(module)
	local source=self:verifyModuleError(module, "source")
	if source==false then return false end
	
	local power=self:verifyNumber(module, "power", 0.5)
	local seamless=self:verifyBoolean(module, "seamless", false)
	
	local f=CGreyscaleErosion()
	f:setPower(power)
	f:setSeamless(seamless)
	f:setGreyscaleSource(source)
	self.moduletable[module.name]=f
	
	print("Module "..module.name.." (erosion,"..power..","..tostring(seamless)..") created.")
	return true
end

function NoiseModuleTree:greyscale_rainfallsim(module)
	local source=self:verifyModuleError(module, "source")
	if source==false then return false end
	
	local seamless=self:verifyBoolean(module, "seamless", false)
	local accumulate=self:verifyBoolean(module, "accumulate", false)
	local numraindrops=self:verifyNumber(module, "num_raindrops", 0)
	
	local f=CGreyscaleRainfallSim()
	f:setSeamless(seamless)
	f:setAccumulate(accumulate)
	f:setNumRaindrops(numraindrops)
	f:setGreyscaleSource(source)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (greyscale_rainfallsim) created.")
	return true
end

	

function NoiseModuleTree:greyscale_fuzzydisk(module)
	local radius=self:verifyNumber(module,"radius",1)
	local centerx=self:verifyNumber(module,"centerx",0)
	local centery=self:verifyNumber(module,"centery",0)
	local seamless=self:verifyBoolean(module,"seamless",false)
	
	local f=CGreyscaleFuzzyDisk()
	f:setRadius(radius)
	f:setCenter(centerx,centery)
	f:setSeamless(seamless)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.."(greyscale_fuzzydisk) created.")
	return true
end

function NoiseModuleTree:greyscale_invert(module)
	local source=self:verifyModuleError(module, "source")
	if source==false then return false end
	
	local f=CGreyscaleInvert()
	f:setGreyscaleSource(source)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.."(greyscale_invert) created.")
	return true
end

function NoiseModuleTree:greyscale_gradient(module)
	local low=self:verifyNumber(module, "low", -1)
	local high=self:verifyNumber(module, "high", 1)
	local direction=self:verifyNumber(module, "direction", 0)
	if direction<0 or direction>1 then
		print("Warning: Gradient direction for module "..module.name.." out of range. Defaulting to 0.")
		direction=0
	end
	
	local f=CGreyscaleGradient()
	f:setType(direction)
	f:setRange(low,high)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (greyscale_gradient) created.")
	return true
end

function NoiseModuleTree:greyscale_remaprange(module)
	local source=self:verifyModuleError(module, "source")
	if source==false then return false end
	
	local low=self:verifyNumber(module, "low", -1)
	local high=self:verifyNumber(module, "high", 1)
	
	local f=CGreyscaleRemapRange()
	f:setGreyscaleSource(source)
	f:setRange(low,high)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (greyscale_remaprange) created.")
	return true
end

function NoiseModuleTree:greyscale_seededvariant(module)
	local source=self:verifyModuleError(module, "source")
	if source==false then return false end
	
	local randomseed=self:verifyNumber(module, "seed", randInt())
	
	local f=CGreyscaleSeededVariant()
	f:setGreyscaleSource(source)
	f:setRandomSeed(randomseed)
	
	self.moduletable[module.name]=f
	
	print("Module "..module.name.." (greyscale_seededvariant) created.")
	return true
end

function NoiseModuleTree:greyscale_turbulence(module)
	local mainsource=self:verifyModuleError(module, "main_source")
	if mainsource==false then return false end
	
	local xsource,ysource,mask
	
	if module.xaxis_source ~= nil then xsource=self.moduletable[module.xaxis_source] end
	if module.yaxis_source ~= nil then ysource=self.moduletable[module.yaxis_source] end
	if module.mask ~= nil then mask=self.moduletable[module.mask] end
	
	local xp,yp=1,1
	
	if xsource then
		xp=self:verifyNumber(module, "xpower", 1)
	end
	
	if ysource then
		yp=self:verifyNumber(module, "ypower", 1)
	end
	
	local f=CGreyscaleTurbulence()
	f:setGreyscaleSource(0,mainsource)
	if xsource then f:setGreyscaleSource(1,xsource) end
	if ysource then f:setGreyscaleSource(2,ysource) end
	if mask then f:setGreyscaleSource(3,mask) end
	f:setPowers(xp,yp)
	
	self.moduletable[module.name]=f
	
	print("Module "..module.name.." (greyscale_turbulence) created.")
	return true
end

function NoiseModuleTree:greyscale_clamp(module)
	local source=self:verifyModuleError(module, "source")
	if source==false then return false end
	
	local low=self:verifyNumber(module, "low", -1)
	local high=self:verifyNumber(module, "high", 1)
	
	local f=CGreyscaleClamp()
	f:setGreyscaleSource(source)
	f:setRange(low,high)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (greyscale_clamp) created.")
	return true
end

function NoiseModuleTree:greyscale_rgbaadapter(module)
	local mainsource=self:verifyModuleError(module, "source")
	if mainsource==false then return false end
	
	local type=self:verifyNumber(module, "mode", 0)
	if type<0 or type>7 then
		print("Warning: Greyscale to RGBA Adapter mode invalid, defaulting to 0.")
		type=0
	end
	
	local f=CGreyscaleRGBAAdapter()
	f:setRGBASource(source)
	f:setMode(type)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (greyscale_rgbaadapter) created.")
	return true
end





----------------------------------------------------------------------------------
-- RGBA Modules

function NoiseModuleTree:rgba_blend(module)
	local source1=self:verifyModuleError(module, "source1")
	local source2=self:verifyModuleError(module, "source2")
	
	if source1==false or source2==false then return false end
	
	local blendop1=self:verifyNumber(module, "blendop1", 0)
	local blendop2=self:verifyNumber(module, "blendop2", 2)
	
	local f=CRGBABlend()
	f:setRGBASource(0,source1)
	f:setRGBASource(1,source2)
	f:setBlendOps(blendop1, blendop2)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (rgba_blend) created.")
	return true
end

function NoiseModuleTree:rgba_blur(module)
	local mainsource=self:verifyModuleError(module, "source")
	if mainsource==false then return false end
	
	local size=self:verifyNumber(module, "size")
	if size==false then return false end
	if size<=0 then print("Error: Blur size <=0") return false end
	
	local seamless=self:verifyBoolean(module, "seamless", false)
	
	local f=CRGBABlur()
	f:setRGBASource(mainsource)
	f:setSize(size)
	f:setSeamless(seamless)
	self.moduletable[module.name]=f
	print("Module "..module.name.." (rgba_blur, "..size..","..seamless..") created.")
	return true
end

function NoiseModuleTree:rgba_colormapadapter(module)
	if module.points==nil or #module.points<2 then
		print("Error: Points array specified for module "..module.name.." nonexistent, or less than 2 entries in length.")
		return false
	end
	
	local source=self:verifyModuleError(module, "source")
	if source==false then return false end
	
	local f=CRGBAColormapAdapter()
	f:setGreyscaleSource(source)
	
	local i
	f:pushPoint(-0.1,module.points[1])
	for i=1,#module.points,1 do
		local v=module.points[i]
		f:pushPoint((i-1)/(#module.points-1),v)
	end
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (rgba_colormap) created.")
	return true
end

function NoiseModuleTree:rgba_combinealpha(module)
	local greysource=self:verifyModuleError(module, "greyscale_source")
	local rgbasource=self:verifyModuleError(module, "rgba_source")
	if greysource==false or rgbasource==false then return false end
	
	local mode=self:verifyNumber(module, "mode", 0)
	if mode<0 or mode >3 then
		print("Warning: mode for module "..module.name.." out of range. Defaulting to 0.")
		mode=0
	end
	
	local f=CRGBACombineAlpha()
	f:setGreyscaleSource(greysource)
	f:setRGBASource(rgbasource)
	f:setMode(mode)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (rgba_combinealpha) created.")
	return true
end

function NoiseModuleTree:rgba_constant(module)
	local r=self:verifyNumber(module, "red", 1)
	local g=self:verifyNumber(module, "green", 1)
	local b=self:verifyNumber(module, "blue", 1)
	local a=self:verifyNumber(module, "alpha", 1)
	
	local f=CRGBAConstant(r,g,b,a)
	self.moduletable[module.name]=f
	
	print("Module "..module.name.." (rgba_constant) created.")
	return true
end

function NoiseModuleTree:rgba_greyscaleadapter(module)
	local source=self:verifyModuleError(module, "source")
	if source==false then return false end
	
	local f=CRGBAAdapterGreyscale()
	f:setGreyscaleSource(source)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (rgba_greyscaleadapter) created.")
	return true
end

function NoiseModuleTree:rgba_highpassfilter(module)
	local source=self:verifyModuleError(module,"source")
	if source==false then return false end
	
	local size=self:verifyNumber(module, "size", 1)
	
	if size<=0 then
		print("Warning: size parameter for module "..module.name.." invalid. Defaulting to 1.")
		size=1
	end
	
	local f=CRGBAHighPassFilter()
	f:setRGBASource(source)
	f:setSize(size)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (rgba_highpassfilter) created.")
	return true
end

function NoiseModuleTree:rgba_image(module)
	local name=self:verifyString(module, "source")
	if name==false then return false end
	
	local f=CRGBAImage()
	f:loadImage(name)
	
	self.moduletable[module.name]=f
	
	print("Module "..module.name.." (rgba_image, "..name..") created.")
	return true
end

function NoiseModuleTree:rgba_invertcolors(module)
	local source=self:verifyModuleError(module, "source")
	if source==false then return false end
	
	local f=CRGBAInvertColors()
	f:setRGBASource(source)
	
	self.moduletable[module.name]=f
	
	print("Module "..module.name.." (rgba_invertcolors) created.")
	return true
end

function NoiseModuleTree:rgba_makeseamless(module)
	local source=self:verifyModuleError(module, "source")
	if source==false then return false end
	
	local seamwidth=self:verifyNumber(module, "seamwidth", 1)
	
	local f=CRGBAMakeSeamless()
	f:setRGBASource(source)
	f:setSeamWidth(seamwidth)
	
	self.moduletable[module.name]=f
	
	return true
end

function NoiseModuleTree:rgba_multrgbagreyscale(module)
	local greysource=self:verifyModuleError(module, "greyscale_source")
	local rgbasource=self:verifyModuleError(module, "rgba_source")
	if greysource==false or rgbasource==false then return false end
	
	local f=CRGBAMultRGBGreyscale()
	f:setGreyscaleSource(greysource)
	f:setRGBASource(rgbasource)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (rgba_multrgbagreyscale) created.")
	return true
end

function NoiseModuleTree:rgba_turbulence(module)
	local mainsource=self:verifyModuleError(module, "main_source")
	if mainsource==false then return false end
	
	local xsource,ysource,mask
	
	if module.xaxis_source ~= nil then xsource=self.moduletable[module.xaxis_source] end
	if module.yaxis_source ~= nil then ysource=self.moduletable[module.yaxis_source] end
	if module.mask ~= nil then mask=self.moduletable[module.mask] end
	
	local xp,yp=1,1
	
	if xsource then
		xp=self:verifyNumber(module, "xpower", 1)
	end
	
	if ysource then
		yp=self:verifyNumber(module, "ypower", 1)
	end
	
	local f=CRGBATurbulence()
	f:setRGBASource(mainsource)
	if xsource then f:setGreyscaleSource(0,xsource) end
	if ysource then f:setGreyscaleSource(1,ysource) end
	if mask then f:setGreyscaleSource(2,mask) end
	f:setPowers(xp,yp)
	
	self.moduletable[module.name]=f
	
	print("Module "..module.name.." (rgba_turbulence) created.")
	return true
end





---------------------------------------------------------------------------------------
-- Image Output Module

function NoiseModuleTree:output(module)
	local source=self:verifyModuleError(module, "source")
	if source==false then return false end
	
	local width=self:verifyNumber(module, "width")
	local height=self:verifyNumber(module, "width")
	local type=self:verifyNumber(module, "mode", 0)
	
	if width==false or height==false then return false end
	if width<=0 or height<=0 then
		print("Error: Output module "..module.name..", invalid image dimensions.")
		return false
	end
	
	local f=COutputModule(width,height,type)
	f:setRGBASource(source)
	
	self.moduletable[module.name]=f
	print("Module "..module.name.." (output) created.")
	return true
end







---------------------------------------------------------------------------------------
-- Frequently Use Combinations

-- These modules can be specified to provide some frequently used module chains, such as fractal turbulence sources and what-not

-- Set up a simple implicit fractal source, and attach a greyscale adapter
function NoiseModuleTree:greyscale_fractal(module)
	local fractal_name=module.name.."_fractal"
	
	local fractal_module={name=fractal_name, type="fractal", fractal_type=module.fractal_type, basis_type=module.basis_type,
		interp_type=module.interp_type, frequency=module.frequency, num_octaves=module.num_octaves}
	
	if self:parseModule(fractal_module)==false then
		print("Error: Could not parse fractal_module.")
		return false
	end
	
	-- Setup an implicit scale/offset module to fix the range
	local ftype=0
	if module.fractal_type=="RIDGEDMULTI" then ftype=1
	elseif module.fractal_type=="BILLOW" then ftype=2
	elseif module.fractal_type=="MULTI" then ftype=3
	elseif module.fractal_type=="HYBRIDMULTI" then ftype=4
	end
	
	local numocts=self:verifyNumber(module, "num_octaves", 1)
	
	-- Setup another scale/offset to remap the range
	
	local remap=self:verifyBoolean(module, "remaprange", false)
	local low=self:verifyNumber(module,"low",-1)
	local high=self:verifyNumber(module,"high",1)
	
	if remap==true then
		local med=low+(high-low)/2
		
		local s=(high-low)/2
		local o=low+s
		local remap_name=module.name.."_remap"
		local remap_module={name=remap_name, type="scaleoffset", source=module.name.."_fractal", scale=s, offset=o}
		if self:parseModule(remap_module)==false then
			print("Error: Couldn't parse remap_module.")
			return false
		end
	end
	
	local rname
	if remap==true then
		rname=module.name.."_remap"
	else
		rname=module.name.."_fractal"
	end
	
	local adapter_name=module.name
	
	local adapter_module={name=adapter_name, type="greyscale_implicitadapter", source=rname,
		mapx1=module.mapx1, mapy1=module.mapy1, mapx2=module.mapx2, mapy2=module.mapy2, mapz1=module.mapz1, mapz2=module.mapz2,
		zdepth=module.zdepth, loopx1=module.loopx1, loopy1=module.loopy1, loopz1=module.loopz1, loopx2=module.loopx2, loopy2=module.loopy2,
		loopz2=module.loopz2, seamlessmode=module.seamlessmode}
	
	if self:parseModule(adapter_module)==false then
		print("Error: Could not parse adapter_module.")
		return false
	end
	
	return true
end

function NoiseModuleTree:greyscale_basisfunction(module)
	local name=module.name
	local basis_name=name.."_basis"
	local basis_module={name=basis_name, type="basisfunction", basis_type=module.basis_type, interp_type=module.interp_type}
	if self:parseModule(basis_module)==false then
		print("Error: Could not parse basis_module.")
		return false
	end
	
	local remap=self:verifyBoolean(module, "remaprange", false)
	
	local adapter_name=module.name
	if remap==true then adapter_name=adapter_name.."_adapter" end
	
	local adapter_module={name=adapter_name, type="greyscale_implicitadapter", source=basis_name,
		mapx1=module.mapx1, mapy1=module.mapy1, mapx2=module.mapx2, mapy2=module.mapy2, mapz1=module.mapz1, mapz2=module.mapz2,
		zdepth=module.zdepth, loopx1=module.loopx1, loopy1=module.loopy1, loopz1=module.loopz1, loopx2=module.loopx2, loopy2=module.loopy2,
		loopz2=module.loopz2, seamlessmode=module.seamlessmode}
		
	if self:parseModule(adapter_module)==false then
		print("Error: Could not parse adapter_module.")
		return false
	end
	
	if remap==true then
		local remap_module={name=module.name, type="greyscale_remaprange", low=module.low, high=module.high, source=adapter_name}
		if self:parseModule(remap_module)==false then
			print("Error: Could not parse remap_module.")
			return false
		end
	end
	
	return true
end

