require 'Class'
dofile("levelgenerator.lua")


-- A word on coordinate frames
-- The WorldContainer maintains a "Current" coordinate frame, representing the center chunk (ie, the current chunk occupied by player or camera
-- controller) and it's neighborhood of neighboring chunks. This is referred to as the "Current" coordinate system. The Current frame is centered at
-- (0,0), meaning that the origin of the center chunk is drawn at (0,0), and neighbors are offset from it in steps of chunksize.

-- Each chunk has an offset from the center chunk origin that defines that chunk's local frame of reference.

-- The "Relative" frame of reference is a set of coordinates specifying GridX/GridY (the integer grid coordinates of a chunk in the world, absolute)
-- and LocalX/LocalY, coordinates relative to the chunk.

-- The WorldContainer provides functions for converting from Relative to Current frames (based on the current center) and from Current to Relative
-- again, based on the current center.

-- Object visual positions should be set to Current coordinate frame each time the object is rendered.

-- Objects should be "owned" by the chunk they reside in (with the player as the sole exception) so that they may be properly streamed to disk when
-- the chunk is unloaded. Objects should try to use Relative coordinates if at all possible, and only convert to current for rendering and for certain
-- calculations (targeting, movement, etc...) to ensure that out-of-date current coords do not get used.


-- NOTE:
-- Object components should properly implement AttachObject and DetachObject to attach/detach objects and visual entities to scene chunks,
-- by querying world_container for a chunk and attaching/detaching themselves and their appropriate visuals.
-- Objects should be detached prior to moving, and reattached once moving is done. They should be visually detached during AddFrameTime, then
-- reattached. 



-- WorldChunk
-- Implements a chunk of world, including the CIsoSceneChunk piece for that section of world, and the list of objects owned by that chunk.
-- Provides functionality to save/load/generate chunks.

WorldChunk=class(function(a,gridx,gridy,chunksize)
	a.gridx=gridx
	a.gridy=gridy
	a.offx=0
	a.offy=0
	a.size=chunksize
	a.isloaded=false
	a.scenechunk=CIsoSceneChunk(0,0,chunksize,chunksize)
	a.lightchunk=CLightChunk(0,0,chunksize+1,chunksize+1)
	a.scenechunk:setLight(a.lightchunk)
	a.objects={}
	--a.generator=LevelGenerator(chunksize,randInt())
	
	--a.generator:test(chunksize)
	
	end)
	
function WorldChunk:setOffset(offx,offy)
	self.offx=offx
	self.offy=offy
	
	-- Set CIsoSceneChunk offset
	self.scenechunk:setOffset(offx,offy)
	self.lightchunk:setOffset(offx,offy)
	
	-- Set all owned entities offsets? Or should entities just own a ptr to parent chunk instead?
end

function WorldChunk:load()
	-- TODO: Implement chunk generating/loading
	
	--console_print("Chunk: "..self.gridx.."x"..self.gridy.." loaded.")
	-- Test, fill with ground
	--local x,y
	--for x=0,self.size-1,1 do
		--for y=0,self.size-1,1 do
			--if(x==y) then self.scenechunk:instanceStaticEntity("basic_ground_plane.mesh", "dirt0001.tga",1,1,0,0, x,0,y,1,1,1,0)
			--else self.scenechunk:instanceStaticEntity("basic_ground_plane.mesh", "testgroundpiece.tga",1,1,0,0, x,0,y,1,1,1,0)
			--end
			
			--if randRange(1,10)<4 then
				--self.lightchunk:setAmbientBlock(x,y)
			--end
		--end
	--end
	
	--for x=1,1000,1 do
		--local rx=randRange(0,self.size-1)
		--local ry=randRange(0,self.size-1)
		--if randRange(1,10)< 5 then
		--self.scenechunk:instanceStaticEntity("characterbillboard.mesh", "turtledude.tga",1,1,0,0, rx,0.01,ry,0.5,0.5,0.5,3)
		--else
		--self.scenechunk:instanceStaticEntity("data/pieces/leftrighttop.msh", "data/tilesets/cliff2.tga", 1, 1, 0, 0, rx,0.01,ry, 0.25,0.25,0.25, 2)
		--end
	--end
	
	-- TODO: Load static light here
	self.lightchunk:setAmbient(1,1,1)
	
	--local c
	--for c=1,5,1 do
		--local rx=randRange(10,self.size-11)
		--local ry=randRange(10,self.size-11)
		--local r=rand01()
		--local g=rand01()
		--local b=rand01()
		--local rad=randRange(5,10)
		--self.lightchunk:addStaticLightPoolCurrentCoords(rx+self.offx, ry+self.offy, rad, r, g, b)
	--end
	generator:generate_chunk(self.scenechunk, self.lightchunk, self.gridx, self.gridy, self.size)
			
	self.isloaded=true
end

function WorldChunk:save()
	-- TODO: Implement chunk saving
	
	if(self.isloaded==false) then
		-- Chunk hasn't been loaded yet, so don't save, just bail
		console_print("Chunk: Can't save, hasn't been loaded.")
		return
	end
	
	-- TODO: Save static light here
	
	console_print("Chunk: "..self.gridx.."x"..self.gridy.." saved.")
	
	-- Dump data prior to chunk going away
	self:killObjects()
	
end

function WorldChunk:attachObject(o)
	local guid=o.guid
	self.objects[guid]=o
end

function WorldChunk:detachObject(o)
	local guid=o.guid
	self.objects[guid]=nil
end

function WorldChunk:attachSceneEntity(entity)
	-- Attach a CIsoSceneEntity
	self.scenechunk:insertEntity(entity)
end

function WorldChunk:detachSceneEntity(entity)
	self.scenechunk:removeEntity(entity)
end

function WorldChunk:killObjects()
	-- Iterate list and kill all objects prior to destroying chunk.
	local i,j
	for i,j in pairs(self.objects) do
		world_container:killObject(j)
	end
end
	
	



WorldContainer=class(function(a, pixelscale, chunksize, neighborhoodsize, start_gridx, start_gridy)
	a.chunks={}
	a.lightchunks={}
	a.center_grid_x=0
	a.center_grid_y=0
	a.chunksize=chunksize
	a.neighborhoodsize=neighborhoodsize
	a.setCenterGrid(a,start_gridx, start_gridy)
	a.view=CIsoSceneView(pixelscale,2)
	a.view:setCameraCenter(0,0)
	a.playerobject=nil
	
	a.nextid=1
	a.objects={}
	
	a.view:debugOutput()
	
	end)
	
function WorldContainer:createObject()
	local o=Object(self.nextid)
	self.objects[self.nextid]=o
	self.nextid = self.nextid+1
	return o
end

function WorldContainer:getObject(guid)
	return self.objects[guid]
end
	
function WorldContainer:reverseProjectOnPlane(mousex, mousey, planey)
	return self.view:reverseProjectOnPlane(mousex, mousey, planey)
end

function WorldContainer:killObject(o)
	local guid=o.guid
	self.objects[guid]=nil
end

function WorldContainer:isChunkLoaded(gridx, gridy)
	-- REturn true if the chunk is loaded and valid
	local gindex=self:gridIndex(gridx,gridy)
	return self:isChunkIndexLoaded(gindex)
end

function WorldContainer:isChunkIndexLoaded(gindex)
	if self.chunks[gindex]==nil then return false end
	if self.chunks[gindex].isloaded==false then return false end
	return true
end


function WorldContainer:getChunk(gridx, gridy)
	local gindex=self:gridIndex(gridx,gridy)
	return self.chunks[gindex]
end

function WorldContainer:getChunkIndex(gindex)
	return self.chunks[gindex]
end
	
function WorldContainer:setCenterGrid(gridx,gridy)
	self.center_grid_x=gridx
	self.center_grid_y=gridy
	
	-- Here, I need to calculate the new Current neighborhood and offsets, stream in new chunks, stream out unneeded chunks, etc.
	self:updateNeighborhood()
end

function WorldContainer:correctRelativeCoords(rel)
	-- Given a set of relative coords, correct them so that overflows in the localx,localy portion are taken away and applied
	-- as modifications to gridx,gridy
	-- In this way, A caller can specify coordinates relative to itself, and have them corrected into the global scope.
	while (rel.localx <0) do
		rel.localx = rel.localx + self.chunksize
		rel.gridx = rel.gridx-1
	end
	while rel.localx >=self.chunksize do
		rel.localx = rel.localx - self.chunksize
		rel.gridx = rel.gridx+1
	end
	while rel.localy <0 do
		rel.localy = rel.localy + self.chunksize
		rel.gridy = rel.gridy-1
	end
	while rel.localy >= self.chunksize do
		rel.localy = rel.localy - self.chunksize
		rel.gridy = rel.gridy+1
	end
	return rel
end

function WorldContainer:isValidCoordRelative(rel)
	-- Is the given coordinate valid, ie is the chunk it points to loaded?
	local gindex=self:gridIndex(rel.gridx, rel.gridy)
	if self.chunks[gindex]==nil then return false else return true end
end
	

function WorldContainer:setCenterGridFromCurrent(current)
	-- Given a set of Current coords (current to the current center) set the new center
	local rel=self:coordsCurrentToRelative(current)
	self:setCenterGrid(rel.gridx, rel.gridy)
	
	-- Should I return update current coords? For now, sure. Maybe I won't need them though.
	return self:coordsRelativeToCurrent(rel)
end

function WorldContainer:setCameraCenterCurrent(cur)
	
	-- Convert to relative, update the chunks, convert back to current and set the view
	local rel=self:coordsCurrentToRelative(cur)
	self:setCenterGrid(rel.gridx, rel.gridy)
	cur=self:coordsRelativeToCurrent(rel)
	self.view:setCameraCenter(cur.x, cur.y)
end

function WorldContainer:setCameraCenterRelative(rel)
	-- Center the world and set the camera relative
	-- Correct coords first
	rel=self:correctRelativeCoords(rel)
	self:setCenterGrid(rel.gridx, rel.gridy)
	local cur=self:coordsRelativeToCurrent(rel)
	self.view:setCameraCenter(cur.x, cur.y)
end

function WorldContainer:gridIndex(gridx, gridy)
	-- Calculate a table index for a grid pair
	return(tostring(gridx).."x"..tostring(gridy))
end

function WorldContainer:coordsRelativeToCurrent(relative)
	-- relative coords consist of a grid/local set
	--local grid_index=self:gridIndex(relative.gridx,relative.gridy)
	--local chunk=self.chunks[grid_index]
	--if(chunk==nil) then
		-- Chunk is not loaded, return nil
		--return nil
	--end
	
	-- Chunk is loaded, so calculate coords
	--return {x=chunk.offx+relative.localx, y=chunk.offy+relative.localy}
	
	
	
	-- Calculate grid coords relative to current center
	local gx=relative.gridx-self.center_grid_x
	local gy=relative.gridy-self.center_grid_y
	
	-- Now, calculate the offset for the relative cell
	local ox=gx*self.chunksize
	local oy=gy*self.chunksize
	
	-- Return coords
	return {x=ox+relative.localx, y=oy+relative.localy}
end

function WorldContainer:coordsCurrentToRelative(current)
	-- Current coords consist of a pair of coords relative to the current Center chunk
	-- Calculate chunk relative to current center
	local cx=current.x/self.chunksize
	local cy=current.y/self.chunksize
	local gx=math.floor(cx)
	local gy=math.floor(cy)
	
	
	-- gx,gy are grid coords relative to center grid, so add to center grid to get absolute grid coords
	return {gridx=gx+self.center_grid_x, gridy=gy+self.center_grid_y, localx=(cx-gx)*self.chunksize, localy=(cy-gy)*self.chunksize}
end


function WorldContainer:calculateNeighborhood(gridx, gridy)
	-- Calculate the neighborhood of the cell
	-- This is dependent upon the neighborhood size parameter which is specified as a radius, greater than 0, around the center grid
	local i,j
	local n={}
	local index=1
	for i=gridx-self.neighborhoodsize,gridx+self.neighborhoodsize,1 do
		for j=gridy-self.neighborhoodsize,gridy+self.neighborhoodsize,1 do
			n[index]={i,j}
			index=index+1
		end
	end
	return n
end

function WorldContainer:isInNeighborhood(n, gridx, gridy)
	local i
	for i=1,#n,1 do
		if gridx==n[i][1] and gridy==n[i][2] then return true end
	end
	return false
end

function WorldContainer:updateNeighborhood()
	-- After setting the center grid, call this to calculate the new neighborhood, update offsets, save and load chunks, etc...
	local n=self:calculateNeighborhood(self.center_grid_x, self.center_grid_y)
	
	
	-- Should I do an update offsets pass to kick it off?
	-- This depends on how I do the loading/saving
	-- If loading and saving take significant amounts of time, I may need to break it up into tasks to be handled one at a time
	-- in between regular updates.
	-- In this case, I need to update offsets of the current set of chunks so that they get rendered correctly
	-- For now, leave this spot blank, but just remember
	
	-- Now, iterate neighborhood and load any new chunks
	local i,j
	for i=1,#n,1 do
		local gx=n[i][1]
		local gy=n[i][2]
		local gindex=self:gridIndex(gx,gy)
		if self.chunks[gindex]==nil then
			-- Build a new chunk
			self.chunks[gindex]=WorldChunk(gx, gy, self.chunksize)
			self.chunks[gindex]:load()
		end
	end
	
	-- Now, iterate chunks, and remove any that aren't in neighborhood
	for i,j in pairs(self.chunks) do
		local gx=j.gridx
		local gy=j.gridy
		if self:isInNeighborhood(n,gx,gy)==false then
			-- Not in neighborhood, save it and remove
			local gindex=self:gridIndex(gx,gy)
			j:save()
			self.chunks[gindex]=nil
		end
	end
	
	-- Now, update offsets
	for i,j in pairs(self.chunks) do
		local gx=j.gridx
		local gy=j.gridy
		local dx=gx-self.center_grid_x
		local dy=gy-self.center_grid_y
		j:setOffset(dx*self.chunksize, dy*self.chunksize)
		
	end
end

function WorldContainer:render()
	self.view:prepForRender()
	for i,j in pairs(self.chunks) do
		self.view:renderChunk(j.scenechunk)
	end
	self.view:render()
end

function WorldContainer:AddFrameTime(elapsedtime, percentwithintick)
	-- AddFrameTime to all active objects, starting with playerobject if set
	local msg={elapsed_time=elapsedtime, percent_within_tick=percentwithintick}
	--if(self.playerobject) then
		--self.playerobject:handleMessage("AddFrameTime", msg)
	--end
	
	-- Now, iterate object list
	local i,j
	for i,j in pairs(self.objects) do
		j:handleMessage("AddFrameTime", msg)
	end
end

function WorldContainer:UpdateLogic()
	-- Update logic
	--if(self.playerobject) then
		--self.playerobject:handleMessage("UpdateLogic")
	--end
	
	local i,j
	for i,j in pairs(self.objects) do
		j:handleMessage("UpdateLogic")
	end
end
